bootstrap: library
from: alpine:3.11.5

%setup
    root=$SINGULARITY_ROOTFS
    rm ${root}/etc/hosts
    rm ${root}/etc/resolv.conf
    ln -s /a/fake/path ${root}/etc/resolv.conf
    ln -s /a/fake/path ${root}/etc/hosts

%post
    echo "TEST OK"