bootstrap: library
from: ubuntu:20.04

%post
  echo "Running %post as $(id -u)"
  # We should appear to be root
  test $(id -u) -eq 0
  # We should be able to install some software
  apt -y update
  apt -y install wget

%test
  echo "Running %post as $(id -u)"
  # We should not appear to be root
  test $(id -u) -ne 0
  # wget was installed
  wget --version
